__version__="0.5.1"
__author__="David J. Wooten"
__citation__="Wooten, DJ and Albert, RA. synergy - A Python library for calculating drug combination synergy. (2020) Bioinformatics. https://doi.org/10.1093/bioinformatics/btaa826"

from . import combination
from . import single
from . import utils